@@ -1,3 +1,5 @@
+import logging
+
 import numpy as np
 from scipy.optimize import minimize
 from sklearn.model_selection import KFold
@@ -6,6 +8,8 @@ from .Kernel import Kernel
 from .Kernel import RBF
 from .Plotting import plot_BayOpt
 
+logger = logging.getLogger(__name__)
+
 
 class GP():
     """
@@ -93,17 +97,17 @@ class GP():
 
         try:
             if self._stat:
-                K=self.get_cov()
+                K = self.get_cov()
             else:
                 K = np.linalg.cholesky(kernel)
 
-            marg = - .5 * Y.T.dot(np.linalg.lstsq(K.T, np.linalg.lstsq(K, Y)[0])[0]) \
+            marg = - .5 * Y.T.dot(np.linalg.lstsq(K.T, np.linalg.lstsq(K, Y, rcond=None)[0], rcond=None)[0]) \
                    - .5 * np.log(np.diag(K)).sum() \
                    - .5 * K.shape[0] * np.log(2 * np.pi)
 
         except np.linalg.LinAlgError as exc:
 
-            print(exc, "\nComputing as a normal inverse\n")
+            logging.info(exc, "\nComputing as a normal inverse\n")
             K = np.linalg.inv(kernel)
             marg = - .5 * Y.T.dot(K.dot(Y)) \
                    - .5 * np.log(np.diag(K)).sum() \
@@ -132,8 +136,8 @@ class GP():
         compute the normal inversion of the Graham matrix. If the set of hyperparameters causes the calculations to fail
         for a np.linalg.error, it sets the marginal likelihood to infinity to discard the set of parameters.
         """
-        if verbose:
-            print(hyper)
+
+        logger.debug(hyper)
 
         if self.normalize_y:
             self._mean_y = np.mean(self.get_Y(), axis=0)
@@ -144,20 +148,21 @@ class GP():
 
         try:
             K = np.linalg.cholesky(kernel)
-            marg = - .5 * Y.T.dot(np.linalg.lstsq(K.T, np.linalg.lstsq(K, Y)[0])[0]) \
+            marg = - .5 * Y.T.dot(np.linalg.lstsq(K.T, np.linalg.lstsq(K, Y, rcond=None)[0], rcond=None)[0]) \
                    - .5 * np.log(np.diag(K)).sum() \
                    - .5 * K.shape[0] * np.log(2 * np.pi)
 
         except np.linalg.LinAlgError as exc:
             try:
-                print(exc, "\nComputing as a normal inverse\n")
+                logger.info(exc, "\nComputing as a normal inverse\n")
                 K = np.linalg.inv(kernel)
                 marg = - .5 * Y.T.dot(K.dot(Y)) \
                        - .5 * np.log(np.diag(K)).sum() \
                        - .5 * K.shape[0] * np.log(2 * np.pi)
 
             except np.linalg.LinAlgError as exc2:
-                print(f"------{exc2}------")
+                if verbose:
+                    logger.warning(exc2)
                 marg = np.inf
 
         if self.get_kernel().get_eval():
@@ -182,20 +187,18 @@ class GP():
         :return: The value of the gradients of negative log marginal likelihood.
         """
 
-        if verbose:
-            print("GRADIENT", hyper)
-
+        logger.debug("GRADIENT", hyper)
         hyper = np.squeeze(hyper)
-        m_g = lambda h:  -(.5 * Y.T.dot(K.dot(h.dot(K.dot(Y)))) - .5 * np.diag(K.dot(h)).sum())
+        m_g = lambda h: -(.5 * Y.T.dot(K.dot(h.dot(K.dot(Y)))) - .5 * np.diag(K.dot(h)).sum())
 
         kernel = self.kernel.kernel_eval_grad_(*hyper, pair_matrix)
         K = np.linalg.inv(kernel[0])
         try:
-            grad_kernel=[m_g(i) for i in kernel[1:]]
+            grad_kernel = [m_g(i) for i in kernel[1:]]
 
         except np.linalg.LinAlgError as exc:
-            print("Numerical Instability\n")
-            raise(exc)
+            if verbose:
+                logger.warning(exc)
 
         return np.array(grad_kernel)
 
@@ -235,8 +238,8 @@ class GP():
 
         for i in hyper_grid:
             tmp_marg = self.compute_log_marginal_likelihood(X, Y, pair_matrix, i, verbose)
-            if check_best(best, tmp_marg):
-                set_dict(tmp_marg, i)
+            if check_best(best, -tmp_marg):
+                set_dict(-tmp_marg, i)
 
         return best
 
@@ -258,10 +261,11 @@ class GP():
         args = (constrains, n_points, function, verbose)
         new = self.grid_search_optimization(*args)
         self.set_marg(new["marg"])
-        self.set_hyper(new["hyper"][0], new["hyper"][1], new["hyper"][2])
+        # fare un buon metodo set hyper
+        self.set_hyper(*new["hyper"])
         self.fit()
 
-    def optimize(self,  n_restarts=10, optimizer="L-BFGS-B", verbose=False):
+    def optimize(self, n_restarts=10, optimizer="L-BFGS-B", verbose=False):
         """
         Optimization Routine for the hyperparameters of the GP by Minimizing the negative log Marginal Likelihood.
         n_restarts : int (default 10)
@@ -281,12 +285,12 @@ class GP():
         self.log_marginal_likelihood()
         old_marg = self.get_marg()
         old_hyper = self.get_kernel().gethyper()
-        hyper_dim=len(old_hyper)
+        hyper_dim = len(old_hyper)
 
-        boundaries=self.get_boundary()
+        boundaries = self.get_boundary()
         eval_grad = self.get_kernel().get_eval()
 
-        print("Starting Log Marginal Likelihood Value: ", old_marg)
+        logger.debug("Starting Log Marginal Likelihood Value: ", old_marg)
 
         X_d, Y_d = self.get_X(), self.get_Y()
         pair_matrix = np.sum(X_d ** 2, axis=1)[:, None] + np.sum(X_d ** 2, axis=1) - 2 * np.dot(X_d, X_d.T)
@@ -295,17 +299,18 @@ class GP():
         new_marg = None
         it = 0
 
-        #Optimization Loop
+        # Optimization Loop
         for i in np.random.uniform(boundaries[:, 0], boundaries[:, 1], size=(n_restarts, hyper_dim)):
             it += 1
-            print("RESTART :", it, i)
+
+            logger.debug("RESTART :", it, i)
 
             res = minimize(lambda h: self.compute_log_marginal_likelihood(X=X_d,
                                                                           Y=Y_d,
                                                                           pair_matrix=pair_matrix,
                                                                           verbose=verbose,
                                                                           hyper=(np.squeeze(
-                                                                                h.reshape(-1, hyper_dim)))),
+                                                                              h.reshape(-1, hyper_dim)))),
                            x0=i,
                            bounds=((1e-5, None), (1e-5, None), (1e-5, None)),
                            method=self.__optimizer,
@@ -314,18 +319,22 @@ class GP():
             if not res.success:
                 continue
 
-            if new_marg is None or res.fun[0] < new_marg:
-                new_marg = res.fun[0]
-                new_hyper = res.x
+            try:
+                if new_marg is None or res.fun[0] < new_marg:
+                    new_marg = res.fun[0]
+                    new_hyper = res.x
+            except:
+                pass
+                logger.info(res.fun)
 
-        #If the optimization doesn't converge set the value to old parameters
+        # If the optimization doesn't converge set the value to old parameters
         if new_marg is None:
-            new_marg=old_marg
-            new_hyper=old_hyper
-
-        print("New Log Marginal Likelihood Value: ", -new_marg)
-        print(new_hyper)
+            if verbose:
+                logger.warning("Using Old Log Marg Likelihood")
+            new_marg = old_marg
+            new_hyper = old_hyper
 
+        logger.info("New Log Marginal Likelihood Value: ", -new_marg, new_hyper)
         # Update and fit the new gp model
         self.set_hyper(*new_hyper)
         self.set_marg(-new_marg)
@@ -345,7 +354,7 @@ class GP():
             self._stat = True
 
         except np.linalg.LinAlgError as exc:
-            raise ValueError("Cholesky decomposition encountered a numerical error\nIncrease the Noise Level\n",exc)
+            raise ValueError("Cholesky decomposition encountered a numerical error\nIncrease the Noise Level\n", exc)
 
     def predict(self, X):
         """
@@ -366,7 +375,7 @@ class GP():
             mean = np.dot(inv_cov.T, np.linalg.solve(self.get_cov(), self.get_Y()))
             # general case
             # Variance
-            #var = ker.gethyper()[0] ** 2 + ker.gethyper()[2] - np.sum(inv_cov ** 2, axis=0)
+            # var = ker.gethyper()[0] ** 2 + ker.gethyper()[2] - np.sum(inv_cov ** 2, axis=0)
             var = ker.kernel_var() - np.sum(inv_cov ** 2, axis=0)
             var_neg = var < 0
             var[var_neg] = 0.
@@ -383,7 +392,7 @@ class GP():
             # DESTANDARDIZE
             y_mean = (self._std_y * y_mean + self._mean_y)
             # VARIANCE
-            #var = ker.gethyper()[0] ** 2 + self.get_noise() - np.sum(inv_cov ** 2, axis=0)
+            # var = ker.gethyper()[0] ** 2 + self.get_noise() - np.sum(inv_cov ** 2, axis=0)
             var = ker.kernel_var() - np.sum(inv_cov ** 2, axis=0)
             # REMOVE VARIANCE VALUE LESS THAN 0
             var_neg = var < 0
@@ -405,7 +414,6 @@ class GP():
         plt = plot_BayOpt(*args)
         plt.show()
 
-
     def prepare_fold(self, n):
         kf = KFold(n_splits=n)
         kf.get_n_splits(self.get_dim_data())
@@ -435,7 +443,6 @@ class GP():
             RMSE += GP.RMSE(cv_gp, self.get_X(), self.get_Y(), index[i][1])
         return RMSE / n_fold
 
-
     def augment_dataset(self, dataset, new_data):
         """
         General function to augment an array with a shape check
@@ -504,7 +511,7 @@ class GP():
             self.__boundary = np.asarray([[1e-4, 10] for i in range(len(self.get_kernel().gethyper()))])
             return self.__boundary
 
-    def set_boundary(self,array):
+    def set_boundary(self, array):
         """
         Create the boundaries for the hyperparameters optimization
         :param array: list
@@ -519,15 +526,15 @@ class GP():
         set_boundary([[1e-5,4]]) on RBF ---> boundaries: (1e-5,4),(1e-5,4),(1e-5,4)
         set_boundary([[1e-5,4],[1,10],[2,3]]) on RBF ---> boundaries: (1e-5,4),(1,10),(2,3)
         """
-        n=len(self.get_kernel().gethyper())
+        n = len(self.get_kernel().gethyper())
 
-        if len(array)==1:
-            self.__boundary=np.asarray([array[0] for i in range(n)])
-        if len(array)==n:
+        if len(array) == 1:
+            self.__boundary = np.asarray([array[0] for i in range(n)])
+        if len(array) == n:
             self.__boundary = np.asarray(array)
 
-
     def set_hyper(self, sigma=None, l=None, noise=None):
+
         self.get_kernel().sethyper(sigma, l, noise)
 
     def get_noise(self):
