@@ -1,3 +1,3 @@
 from .GP import GP, generate_grid
-from .util import time_log, Observer
+from .util import time_log, Observer, log_gp
 from .Plotting import plot_BayOpt
\ No newline at end of file
